{-# LANGUAGE QuasiQuotes #-}
module Promela.Declarations where

import SCons
import Promela.VarNames
import Promela.Promela
import Promela.FromFile
import Text.Printf (printf)
import Data.List (intercalate, elemIndex, nub)
import Data.Maybe (fromJust)
import qualified Data.Text as T

data FDesc = FDesc {
  fd_name :: FName
  , fd_cntVals :: FVal
  }
  deriving (Eq, Show)

type StDesc = [FDesc]

algoFields :: Algo -> [FName]
algoFields a = (if a_type a == Timestamped then ["ts"] else []) ++ 
             concatMap (\s -> sendField s : map updTarget (instrs s)) (a_rounds a)

toDescription :: Algo -> StDesc
toDescription a = map descField $ nub $ algoFields a
  where
    n = cutoff a
    descField fName
      | fName == inpName    = FDesc inpName 2
      | fName == "ts"       = FDesc fName n
      | otherwise           = FDesc fName 3
    

genStEncoder :: StDesc -> String
genStEncoder desc = pMacro encoderName fNames expr
  where
    fNames = map fd_name desc
    fCnts = map fd_cntVals desc
    params = intercalate ", " fNames
    expr = "(" ++ aux desc ++ ")"
    aux [] = "0"
    aux ((FDesc fn fc):fds) = printf "%s + %d * (%s)" fn fc (aux fds)

genStDecoder :: StDesc -> FName -> String
genStDecoder desc f = pMacro (decoderName f) ["st"] expr
  where
    expr = aux "st" desc
    aux :: String -> StDesc -> String
    aux _ [] = error "Field not present in the description"
    aux v ((FDesc f' fc):fds)
      | f' == f     = printf "((%s) %% %d)" v fc
      | otherwise   = aux (printf "(%s / %d)" v fc) fds

genStDecoders :: StDesc -> String
genStDecoders desc = unlines $ map (genStDecoder desc) (map fd_name desc)

genStUpdater :: StDesc -> FName -> String
genStUpdater desc f = printf "#define %s(st, val) encode_st(%s)" (updaterName f) expr
  where
    fNames = map fd_name desc
    expr = intercalate ", " updExprs
    updExprs = map updExpr fNames
    updExpr f'
      | f == f'     = "val"
      | otherwise   = printf "%s(st)" (decoderName f')

genStUpdaters :: StDesc -> String
genStUpdaters desc = unlines $ map (genStUpdater desc) (map fd_name desc)

nrStates :: StDesc -> Int
nrStates desc = product $ map fd_cntVals desc

varType :: Int -> String
varType n
  | n < 256     = "byte"
  | otherwise   = "short"

genStDecl :: StDesc -> String
genStDecl desc = printf "#define %s %d\n\
                        \byte %s[%s];"
                 cntBoundName (nrStates desc)
                 ctrVarName cntBoundName
    

genInitial :: StDesc -> String
genInitial desc = pMacro "INITIAL" [] body
  where
    body = pCall encoderName initVals
    initVals = map (\fName -> if fName `elem` [inpName, "ts"] then "0" else "BOT")
               $ map fd_name desc

testDesc :: StDesc
testDesc = [FDesc "inp" 2, FDesc "eph" 3, FDesc "dec" 3, FDesc "ts" 5]

generateDeclarations :: Algo -> T.Text
generateDeclarations a = T.pack $ unlines [
  defineConstants
  , updateMacros
  , genStEncoder desc
  , genInitial desc
  , genStDecoders desc
  , genStUpdaters desc
  , genStDecl desc
  , if hasLdr a then pDeclare (varType (nrStates desc)) ldrStVarName else ""
  , rndAndUpdDecls
  , auxVarsDecls
  , tsInlineDecls
  , phiVarsDecls
  , startVarDecl
  , specVarsDecls
    ]
  where
    desc = toDescription a
    defineConstants = unlines [
      pMacro "N" [] (show (cutoff a))
      , pMacro "BOT" [] "2"
      , pMacro "NEG_BYTE" [] "255"
      ]
    updateMacros = unlines [
      pMacro "_pupds1" ["i"] "poss_upds[i]"
      , pMacro "_pupds2" ["i", "j"] "poss_upds[i * 3 + j]"
      , pMacro "_topu1" ["x"] "((x)%3)"
      , pMacro "_topu12" ["x"] "(((x)%9)/3)"
      , pMacro "_topu22" ["x"] "(((x)%9)%3)"
      , pMacro "_enabled_pu" ["x"] "_pupds1((x)%3)"
      , pMacro "_enabled_pu2" ["x"] "_pupds1((x)%9)"
      , pMacro "_bot_pu1" ["x"] "(_topu1(x) == 2)"
      , pMacro "_bot_pu12" ["x"] "(_topu12(x) == 2)"
      , pMacro "_bot_pu22" ["x"] "(_topu22(x) == 2)"
      ]
    declareVars = unlines . map (uncurry pDeclare)
    rndAndUpdDecls = declareVars [
      ("byte", rndVar)
      , ("byte", pArr "poss_upds" "3 * 3")
      , ("byte", pArr updChoiceName "N")
      ]
    auxVarsDecls =  declareVars [
      (varType (nrStates desc), pArr newStName "N")
      , ("byte", iVar 1)
      , ("byte", iVar 2)
      , ("byte", iVar 3)
      , ("byte", iVar 4)
      , ("byte", nVar 0)
      , ("byte", nVar 1)
      ] ++ (if isTSed a then tsVarsDecls else [])
        ++ (if isRand a then randVarsDecls else [])
    tsVarsDecls = declareVars [
      ("bit", compactName)
      , ("byte", "diff_to_next_ts = NEG_BYTE") -- Ugh. TODO: handle intialization as well?
      , ("byte", pArr "ts_of_0" "N")
      , ("byte", pArr "ts_of_1" "N")
      , ("byte", pArr "ts" "N")
      , ("byte", "swap_tmp")
      ]
    tsInlineDecls
      | isTSed a    = T.unpack [fromFile|src/Promela/fragments/swap.pml|]
      | otherwise   = ""
    randVarsDecls = declareVars [
          ("bit", pArr "urand" "N")
          , ("bit", "phi_lucky")
          ]
    phiCnt = length $ a_term a
    phiVarsDecls = declareVars $ map (\i -> ("bit", phiName i)) [1..phiCnt]
    specVarsDecls = declareVars [
      ("bit", "ex_d0")
      , ("bit", "ex_d1")
      ]
    startVarDecl
      | phiCnt > 1  = declareVars [("bit", "start")]
      | otherwise   = ""

