module SCons.Parser (parseAlgo) where

import SCons
import Text.Parsec
import Text.Parsec.Token
import Text.Parsec.Language
import qualified Data.Text as T
import Data.Ratio ((%))

type Parser a = Parsec String () a

runSP :: Parser a -> String -> Either ParseError a
runSP p = runParser p () ""

def = emptyDef { 
  commentLine = "#"
  , identStart = letter
  , identLetter = alphaNum
  , opLetter = oneOf "/>"
  , reservedOpNames = ["/", ">"]
  , reservedNames = ["A2A", "L2A", "A2L", "TERM", "SAFETY"
                    , "Any", "Smin", "Smor", "SmaxTS", "AllEq"
                    , "PhiLS", "PhiLR", "PhiT", "PhiU"
                    , "NORMAL", "TIMESTAMPED", "RANDOMIZED"
                    ]
  }
  
TokenParser {
  parens = t_parens
  , reservedOp = t_reservedOp
  , reserved = t_reserved
  , semi = t_semi
  , comma = t_comma
  , commaSep = t_commaSep
  , semiSep = t_semiSep
  , natural = t_natural
  , identifier = t_identifier
  } = makeTokenParser def

t_predicate :: Parser Predicate
t_predicate = foldr1 (<|>) [
  (t_reserved "Any" >> return Any)
  , (t_reserved "Smin" >> return Smin)
  , (t_reserved "Smor" >> return Smor)
  , (t_reserved "SmaxTS" >> return SmaxTS)
  , (t_reserved "AllEq" >> return AllEq)
  ]
 
t_cp :: Parser CommPattern
t_cp = foldr1 (<|>) [
  (t_reserved "A2A" >> return AllToAll)
  , (t_reserved "A2L" >> return AllToLdr)
  , (t_reserved "L2A" >> return LdrToAll)
  ]
 
t_commFormula :: Parser CommFormula
t_commFormula = foldr1 (<|>) [
  (t_reserved "PhiLR" >> rational >>= (return . PhiLR))
  , (t_reserved "PhiLS" >> return PhiLS)
  , (t_reserved "PhiT" >> rational >>= (return . PhiT))
  , (t_reserved "PhiU" >> return PhiU)
  ]


t_type :: Parser AlgoType
t_type = foldr1 (<|>) [
  t_reserved "NORMAL" >> return Normal
  , t_reserved "TIMESTAMPED" >> return Timestamped
  , t_reserved "RANDOMIZED" >> return Randomized
  ]

rational :: Parser Rational
rational = do
  x <- t_natural
  y <- rest <|> return 1
  if y /= 0
    then return (x % y)
    else fail "Denumerator is 0" 
  where
    rest = do
      t_reservedOp "/"
      t_natural

guard :: Parser (Rational, Predicate)
guard = t_parens $ do
  th <- rational
  t_comma
  p <- t_predicate
  return (th, p)
  
updInstr :: Parser UpdInstr
updInstr = do
  (th, p) <- guard
  t_reservedOp ">"
  field <- t_identifier
  return $ Upd th p field
  
send :: Parser (String, CommPattern)
send = do
  field <- t_identifier
  cp <- t_cp
  return (field, cp)
  
rspec :: Parser RSpec
rspec = do
  (sField, cp) <- send
  instrs <- many updInstr
  return $ RSpec sField cp instrs

rFormula = t_commaSep t_commFormula

term :: Parser [RFormula]
term = do
  t_reserved "TERM"
  t_semiSep rFormula
  
safety :: Parser [RFormula]
safety = do
  t_reserved "SAFETY"
  t_semiSep rFormula
 
algo :: Parser Algo
algo = do
  typ <- t_type
  rnds <- many1 rspec
  let rCnt = length rnds
  safety <- option (replicate rCnt []) safety
  terms <- option [] $ many1 term
  return $ Algo typ rnds safety terms
  
parseAlgo :: String -> Either ParseError Algo
parseAlgo = runSP algo

   
